@@ -4,16 +4,17 @@ from concurrent.futures import ThreadPoolExecutor
 
 import certifi
 import urllib3
-from tqdm import tqdm
-
 from picomc.env import Env
 from picomc.logging import logger
+from tqdm import tqdm
 
 
 def downloader_urllib3(q, d, workers=8):
     http_pool = urllib3.PoolManager(cert_reqs="CERT_REQUIRED", ca_certs=certifi.where())
     total = len(q)
 
+    errors = []
+
     def dl(i, url, outs):
         aouts = list(os.path.join(d, o) for o in outs)
         for o in aouts:
@@ -21,6 +22,12 @@ def downloader_urllib3(q, d, workers=8):
         pout, *eouts = aouts
         logger.debug("Downloading [{}/{}]: {}".format(i, total, url))
         resp = http_pool.request("GET", url, preload_content=False)
+        if resp.status != 200:
+            errors.append(
+                "Failed to download ({}) [{}/{}]: {}".format(resp.status, i, total, url)
+            )
+            resp.release_conn()
+            return
         with open(pout, "wb") as poutfd:
             shutil.copyfileobj(resp, poutfd)
         resp.release_conn()
@@ -41,6 +48,11 @@ def downloader_urllib3(q, d, workers=8):
             fut = tpe.submit(dl, i, url, outs)
             fut.add_done_callback(done)
 
+    for error in errors:
+        logger.warn(error)
+
+    return not errors
+
 
 class DownloadQueue:
     def __init__(self):
@@ -54,8 +66,8 @@ class DownloadQueue:
 
     def download(self, d):
         if not self.q:
-            return
+            return True
         else:
-            logger.info("Using parallel urllib3 downloader.")
+            logger.debug("Using parallel urllib3 downloader.")
             downloader = downloader_urllib3
-        downloader(self.q, d)
+        return downloader(self.q, d)
