@@ -4,7 +4,7 @@ import sys
 from functools import partial
 
 from picomc.env import get_filepath
-from picomc.javainfo import java_info, java_version
+from picomc.javainfo import get_java_info, get_java_version
 from picomc.logging import logger
 
 
@@ -76,12 +76,12 @@ def check_directories():
 
 def assert_java(java):
     try:
-        jinfo = java_info(java)
-        jver = java_version(java)
+        jinfo = get_java_info(java)
+        jver = get_java_version(java)
         badjv = False
         if jinfo:
-            badjv = not jinfo["java.version"].startswith("1.8.0")
-            bitness = jinfo.get("sun.arch.data.model", None)
+            badjv = not jinfo["java.version"].decode("ascii").startswith("1.8.0")
+            bitness = jinfo.get("sun.arch.data.model", None).decode("ascii")
             if bitness and bitness != "64":
                 logger.warn(
                     "You are not using 64-bit java. Things will probably not work."
