@@ -1,22 +1,130 @@
 import json
-import operator
 import os
 import posixpath
 import shutil
 import urllib.parse
 import urllib.request
+from dataclasses import dataclass
 from functools import reduce
 from platform import architecture
 from string import Template
 
 import requests
-
 from picomc.downloader import DownloadQueue
 from picomc.env import Env, get_filepath
+from picomc.javainfo import get_java_info
 from picomc.logging import logger
+from picomc.rules import match_ruleset
 from picomc.utils import cached_property, die, file_sha1, file_verify_relative
 
 
+@dataclass
+class LibraryArtifact:
+    url: str
+    path: str
+    sha1: str
+    size: int
+    filename: str
+
+    @classmethod
+    def from_json(cls, obj):
+        return cls(
+            url=obj["url"],
+            path=obj["path"],
+            sha1=obj["sha1"],
+            size=obj["size"],
+            filename=None,
+        )
+
+
+class Library:
+    MOJANG_BASE_URL = "https://libraries.minecraft.net/"
+
+    def __init__(self, json_lib):
+        self.json_lib = json_lib
+        self._populate()
+
+    def _populate(self):
+        js = self.json_lib
+        self.libname = js["name"]
+        self.is_native = "natives" in js
+        self.base_url = js.get("url", Library.MOJANG_BASE_URL)
+
+        self.available = True
+
+        self.native_suffix = ""
+        if self.is_native:
+            try:
+                classifier_tmpl = self.json_lib["natives"][Env.platform]
+                arch = architecture()[0][:2]
+                self.native_classifier = Template(classifier_tmpl).substitute(arch=arch)
+                self.native_suffix = "-" + self.native_classifier
+            except KeyError:
+                logger.warn(
+                    f"Native {self.libname} is not available for current platform {Env.platform}."
+                )
+                self.native_classifier = None
+                self.available = False
+                return
+
+        self.virt_artifact = self.make_virtual_artifact()
+        self.artifact = self.resolve_artifact()
+
+        # Just use filename and path derived from the name.
+        self.filename = self.virt_artifact.url
+        self.path = self.virt_artifact.path
+
+        # Actual fs path
+        self.relpath = os.path.join(*self.path.split("/"))
+
+        if self.artifact:
+            final_art = self.artifact
+
+            # Sanity check
+            assert self.virt_artifact.path == self.artifact.path
+        else:
+            final_art = self.virt_artifact
+
+        self.url = final_art.url
+        self.sha1 = final_art.sha1
+        self.size = final_art.size
+
+    def make_virtual_artifact(self):
+        # I don't know where the *va part comes from, it was already implemented
+        # before a refactor, unfortunately the reason was not documented.
+        # Currently I don't have a version in my versions directory which
+        # utilizes that.
+        # I am leaving it implemented, as there probably was motivation to do it.
+        group, art_id, version, *va = self.libname.split(":")
+        group = group.replace(".", "/")
+        v2 = "-".join([version] + va)
+
+        filename = f"{art_id}-{v2}{self.native_suffix}.jar"
+        path = f"{group}/{art_id}/{version}/{filename}"
+        url = urllib.parse.urljoin(self.base_url, path)
+
+        return LibraryArtifact(
+            url=url, path=path, sha1=None, size=None, filename=filename
+        )
+
+    def resolve_artifact(self):
+        if self.is_native:
+            if self.native_classifier is None:
+                # Native not available for current platform
+                return None
+            else:
+                art = self.json_lib["downloads"]["classifiers"][self.native_classifier]
+                return LibraryArtifact.from_json(art)
+        else:
+            try:
+                return LibraryArtifact.from_json(self.json_lib["downloads"]["artifact"])
+            except KeyError:
+                return None
+
+    def get_abspath(self, library_root):
+        return os.path.join(library_root, self.relpath)
+
+
 class VersionType:
     MAP = {"release": 1, "snapshot": 2, "old_alpha": 4, "old_beta": 8}
 
@@ -116,7 +224,9 @@ class VersionSpec:
     arguments = CachedVspecAttr("arguments", mode=MODE_REDUCE, reduce_func=argumentadd)
     mainClass = CachedVspecAttr("mainClass")
     assetIndex = CachedVspecAttr("assetIndex")
-    libraries = CachedVspecAttr("libraries", mode=MODE_REDUCE, reduce_func=operator.add)
+    libraries = CachedVspecAttr(
+        "libraries", mode=MODE_REDUCE, reduce_func=lambda x, y: y + x
+    )
     jar = CachedVspecAttr("jar", default=lambda vobj: vobj.version_name)
     downloads = CachedVspecAttr("downloads", default=lambda vobj: {})
 
@@ -127,6 +237,7 @@ class Version:
 
     def __init__(self, version_name):
         self.version_name = version_name
+        self._libraries = dict()
 
     @cached_property
     def jarfile(self):
@@ -197,93 +308,19 @@ class Version:
         except requests.ConnectionError:
             die("Failed to retrieve asset index.")
 
-    def _libraries(self, natives_only=False):
-        # The rule matching in this function could be cached,
-        # not sure if worth it.
-        for lib in self.vspec.libraries:
-            rules = lib.get("rules", [])
-            allow = "rules" not in lib
-            for rule in rules:
-                osmatch = True
-                if "os" in rule:
-                    osmatch = rule["os"]["name"] == Env.platform
-                if osmatch:
-                    allow = rule["action"] == "allow"
-            if not allow:
-                continue
-            if natives_only and "natives" not in lib:
-                continue
-            yield lib
-
-    @staticmethod
-    def _resolve_library(lib):
-        # TODO
-        # For some reason I don't remember, we are constructing the paths
-        # to library downloads manually instead of using the url and hash
-        # provided in the vspec. This should probably be reworked and hashes
-        # should be checked instead of just the filenames.
-        #
-        # The reason is that the downloads tag is only present in vanilla vspec,
-        # forge, optifine, fabric don't provide it.
-        suffix = ""
-        sha = None
-        if "natives" in lib:
-            if Env.platform in lib["natives"]:
-                suffix = "-" + lib["natives"][Env.platform]
-                suffix = Template(suffix).substitute(arch=architecture()[0][:2])
-                try:
-                    sha = lib["downloads"]["classifiers"]["natives-" + Env.platform][
-                        "sha1"
-                    ]
-                except KeyError:
-                    pass
-            else:
-                logger.warn(
-                    (
-                        "Native library ({}) not available"
-                        "for current platform ({}). Ignoring."
-                    ).format(lib["name"], Env.platform)
-                )
-                return None
+    def get_libraries(self, java_info):
+        key = java_info.get("java.home", None)
+        if key and key in self._libraries:
+            return self._libraries[key]
         else:
-            try:
-                sha = lib["downloads"]["artifact"]["sha1"]
-            except KeyError:
-                pass
-
-        if sha is None:
-            logger.debug("Library {} has no sha in vspec".format(lib["name"]))
-
-        fullname = lib["name"]
-        url_base = lib.get("url", Version.LIBRARIES_URL)
-        p, n, v, *va = fullname.split(":")
-        v2 = "-".join([v] + va)
-
-        # TODO this is fugly
-        class LibPaths:
-            package = p.replace(".", "/")
-            name = n
-            version = v
-            ext_version = v2
-            filename = "{}-{}{}.jar".format(name, ext_version, suffix)
-            fullpath = "{}/{}/{}/{}".format(package, name, version, filename)
-            url = urllib.parse.urljoin(url_base, fullpath)
-            basedir = get_filepath("libraries")
-            local_relpath = os.path.join(*fullpath.split("/"))
-            local_abspath = os.path.join(basedir, local_relpath)
-
-        LibPaths.sha = sha
-
-        return LibPaths
-
-    def lib_filenames(self, natives=False):
-        for lib in self._libraries(natives):
-            if not natives and "natives" in lib:
-                continue
-            paths = self._resolve_library(lib)
-            if not paths:
-                continue
-            yield paths.local_abspath
+            libs = []
+            for lib in self.vspec.libraries:
+                if "rules" in lib and not match_ruleset(lib["rules"], java_info):
+                    continue
+                libs.append(Library(lib))
+            if key:
+                self._libraries[key] = libs
+            return libs
 
     def download_jarfile(self, force=False):
         """Checks existence and hash of cached jar. Downloads a new one
@@ -306,23 +343,31 @@ class Version:
             logger.info("Downloading jar ({}).".format(self.version_name))
             urllib.request.urlretrieve(dlspec["url"], self.jarfile)
 
-    def download_libraries(self, force=False):
+    def download_libraries(self, java_info, force=False):
         """Downloads missing libraries."""
         logger.info("Checking libraries.")
         q = DownloadQueue()
-        for lib in self._libraries():
-            paths = self._resolve_library(lib)
-            if not paths:
+        for library in self.get_libraries(java_info):
+            if not library.available:
+                continue
+            basedir = get_filepath("libraries")
+            abspath = library.get_abspath(basedir)
+            ok = os.path.isfile(abspath) and os.path.getsize(abspath) > 0
+            if library.sha1 is not None:
+                ok = ok and file_sha1(abspath) == library.sha1
+            if not ok and not library.url:
+                logger.error(
+                    f"Library {library.libname} is missing or corrupt and has no download url."
+                )
                 continue
-            ok = (
-                os.path.isfile(paths.local_abspath)
-                and os.path.getsize(paths.local_abspath) > 0
-            )
-            if paths.sha is not None:
-                ok = ok and file_sha1(paths.local_abspath) == paths.sha
             if force or not ok:
-                q.add(paths.url, paths.local_relpath)
-        q.download(paths.basedir)
+                q.add(library.url, library.relpath)
+        if len(q) > 0:
+            logger.info("Downloading {} libraries.".format(len(q)))
+        if not q.download(basedir):
+            logger.error(
+                "Some libraries failed to download. If they are part of a non-vanilla profile, the original installer may need to be used."
+            )
 
     def _populate_virtual_assets(self, where):
         for name, obj in self.raw_asset_index["objects"].items():
@@ -363,8 +408,10 @@ class Version:
             url = urllib.parse.urljoin(self.ASSETS_URL, posixpath.join(sha[0:2], sha))
             q.add(url, path)
 
-        logger.info("Downloading {} assets. This could take a while.".format(len(q)))
-        q.download(Env.app_root)
+        if len(q) > 0:
+            logger.info("Downloading {} assets.".format(len(q)))
+        if not q.download(Env.app_root):
+            logger.error("Some assets failed to download.")
 
         if is_virtual:
             logger.info("Copying virtual assets")
@@ -372,13 +419,15 @@ class Version:
             logger.debug("Virtual asset path: {}".format(where))
             self._populate_virtual_assets(where)
 
-    def prepare(self):
+    def prepare(self, java_info=None):
+        if not java_info:
+            java_info = get_java_info(Env.gconf.get("java.path"))
         self.download_jarfile()
-        self.download_libraries()
+        self.download_libraries(java_info)
         self.download_assets()
 
-    def prepare_launch(self, gamedir):
-        self.prepare()
+    def prepare_launch(self, gamedir, java_info):
+        self.prepare(java_info)
         self.prepare_assets_launch(gamedir)
 
 
@@ -391,10 +440,10 @@ class VersionManager:
         """Takes a metaversion and resolves to a version."""
         if v == "latest":
             v = self.manifest["latest"]["release"]
-            logger.info("Resolved latest -> {}".format(v))
+            logger.debug("Resolved latest -> {}".format(v))
         elif v == "snapshot":
             v = self.manifest["latest"]["snapshot"]
-            logger.info("Resolved snapshot -> {}".format(v))
+            logger.debug("Resolved snapshot -> {}".format(v))
         return v
 
     @cached_property
